@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -32,9 +32,12 @@ extern "C" {
  * @short_description: Functions for handling utsname
  */
 
+const char *connman_utsname_get_hostname(void);
+
 struct connman_utsname_driver {
 	const char *name;
 	int priority;
+	const char * (*get_hostname) (void);
 	int (*set_hostname) (const char *hostname);
 	int (*set_domainname) (const char *domainname);
 };
