@@ -72,6 +72,7 @@ struct connman_config_service {
 	char *ipv6_gateway;
 	char *ipv6_privacy;
 	char *mac;
+	char *devname;
 	bool mdns;
 	char **nameservers;
 	char **search_domains;
@@ -119,6 +120,7 @@ static bool cleanup = false;
 #define SERVICE_KEY_IPv6               "IPv6"
 #define SERVICE_KEY_IPv6_PRIVACY       "IPv6.Privacy"
 #define SERVICE_KEY_MAC                "MAC"
+#define SERVICE_KEY_DEVICE_NAME	       "DeviceName"
 #define SERVICE_KEY_NAMESERVERS        "Nameservers"
 #define SERVICE_KEY_SEARCH_DOMAINS     "SearchDomains"
 #define SERVICE_KEY_TIMESERVERS        "Timeservers"
@@ -154,6 +156,7 @@ static const char *service_possible_keys[] = {
 	SERVICE_KEY_IPv6,
 	SERVICE_KEY_IPv6_PRIVACY,
 	SERVICE_KEY_MAC,
+	SERVICE_KEY_DEVICE_NAME,
 	SERVICE_KEY_MDNS,
 	SERVICE_KEY_NAMESERVERS,
 	SERVICE_KEY_SEARCH_DOMAINS,
@@ -196,7 +199,7 @@ static void unregister_service(gpointer data)
 							list = list->next) {
 		service_id = list->data;
 
-		service = __connman_service_lookup_from_ident(service_id);
+		service = connman_service_lookup_from_identifier(service_id);
 		if (service) {
 			__connman_service_set_immutable(service, false);
 			__connman_service_set_config(service, NULL, NULL);
@@ -257,6 +260,7 @@ free_only:
 	g_free(config_service->ipv6_gateway);
 	g_free(config_service->ipv6_privacy);
 	g_free(config_service->mac);
+	g_free(config_service->devname);
 	g_strfreev(config_service->nameservers);
 	g_strfreev(config_service->search_domains);
 	g_strfreev(config_service->timeservers);
@@ -478,6 +482,12 @@ static bool load_service_generic(GKeyFile *keyfile,
 		service->mac = str;
 	}
 
+	str = __connman_config_get_string(keyfile, group, SERVICE_KEY_DEVICE_NAME, NULL);
+	if (str) {
+		g_free(service->devname);
+		service->devname = str;
+	}
+
 	str = __connman_config_get_string(keyfile, group, SERVICE_KEY_DOMAIN, NULL);
 	if (str) {
 		g_free(service->domain_name);
@@ -531,6 +541,7 @@ err:
 	g_free(service->ipv6_address);
 	g_free(service->ipv6_gateway);
 	g_free(service->mac);
+	g_free(service->devname);
 	g_free(service);
 
 	return false;
@@ -1095,8 +1106,7 @@ static char *config_pem_fsid(const char *pem_file)
 
 static void provision_service_wifi(struct connman_config_service *config,
 				struct connman_service *service,
-				struct connman_network *network,
-				const void *ssid, unsigned int ssid_len)
+				struct connman_network *network)
 {
 	if (config->eap)
 		__connman_service_set_string(service, "EAP", config->eap);
@@ -1253,7 +1263,7 @@ static int try_provision_service(struct connman_config_service *config,
 	}
 
 	DBG("service %p ident %s", service,
-					__connman_service_get_ident(service));
+				connman_service_get_identifier(service));
 
 	if (config->mac) {
 		struct connman_device *device;
@@ -1271,6 +1281,22 @@ static int try_provision_service(struct connman_config_service *config,
 
 		if (g_ascii_strcasecmp(device_addr, config->mac) != 0)
 			return -ENOENT;
+	} else if (config->devname) {
+		struct connman_device *device;
+		const char *devname;
+
+		device = connman_network_get_device(network);
+		if (!device) {
+			connman_error("Network device is missing");
+			return -ENODEV;
+		}
+
+		devname = connman_device_get_string(device, "Interface");
+
+		DBG("wants %s has %s", config->devname, devname);
+
+		if (g_ascii_strcasecmp(devname, config->devname) != 0)
+			return -ENOENT;
 	}
 
 	if (!config->ipv6_address) {
@@ -1356,7 +1382,7 @@ static int try_provision_service(struct connman_config_service *config,
 
 	__connman_service_disconnect(service);
 
-	service_id = __connman_service_get_ident(service);
+	service_id = connman_service_get_identifier(service);
 	config->service_identifiers =
 		g_slist_prepend(config->service_identifiers,
 				g_strdup(service_id));
@@ -1391,8 +1417,7 @@ static int try_provision_service(struct connman_config_service *config,
 						config->timeservers);
 
 	if (type == CONNMAN_SERVICE_TYPE_WIFI) {
-		provision_service_wifi(config, service, network,
-							ssid, ssid_len);
+		provision_service_wifi(config, service, network);
 	}
 
 	__connman_service_mark_dirty();
@@ -1703,7 +1728,6 @@ void connman_config_free_entries(struct connman_config_entry **entries)
 	}
 
 	g_free(entries);
-	return;
 }
 
 bool __connman_config_address_provisioned(const char *address,
