@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2013  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -67,11 +67,11 @@ static void detect_newlink(unsigned short type, int index,
 	}
 
 	device = find_device(index);
-	if (device != NULL)
+	if (device)
 		return;
 
 	device = connman_device_create_from_index(index);
-	if (device == NULL)
+	if (!device)
 		return;
 
 	if (connman_device_register(device) < 0) {
@@ -90,7 +90,7 @@ static void detect_dellink(unsigned short type, int index,
 	DBG("type %d index %d", type, index);
 
 	device = find_device(index);
-	if (device == NULL)
+	if (!device)
 		return;
 
 	device_list = g_slist_remove(device_list, device);
