@@ -39,6 +39,7 @@ DBusMessage *__connman_error_failed(DBusMessage *msg, int errnum)
 		return __connman_error_not_registered(msg);
 	case ENXIO:
 		return __connman_error_not_found(msg);
+	case EPERM:
 	case EACCES:
 		return __connman_error_permission_denied(msg);
 	case EEXIST:
@@ -67,6 +68,8 @@ DBusMessage *__connman_error_failed(DBusMessage *msg, int errnum)
 		return __connman_error_in_progress(msg);
 	case ENOKEY:
 		return __connman_error_passphrase_required(msg);
+	case ECANCELED:
+		return __connman_error_operation_canceled(msg);
 	}
 
 	return g_dbus_create_error(msg, CONNMAN_ERROR_INTERFACE
@@ -185,3 +188,9 @@ DBusMessage *__connman_error_invalid_property(DBusMessage *msg)
 	return g_dbus_create_error(msg, CONNMAN_ERROR_INTERFACE
 				".InvalidProperty", "Invalid property");
 }
+
+DBusMessage *__connman_error_operation_canceled(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, CONNMAN_ERROR_INTERFACE
+				".OperationCanceled", "Operation canceled");
+}
