@@ -66,7 +66,7 @@ static void ntp_callback(bool success, void *user_data)
 		return;
 	}
 
-	if (!gettimeofday(&tv, NULL)) {
+	if (gettimeofday(&tv, NULL) < 0) {
 		connman_warn("Failed to get current time");
 	}
 
@@ -398,7 +398,7 @@ static void ts_reset(struct connman_service *service)
 
 void __connman_timeserver_sync(struct connman_service *service)
 {
-	if (!service || service == ts_service)
+	if (!service || ts_service == service)
 		return;
 
 	ts_reset(service);
@@ -406,7 +406,7 @@ void __connman_timeserver_sync(struct connman_service *service)
 
 void __connman_timeserver_conf_update(struct connman_service *service)
 {
-	if (!service || service != ts_service)
+	if (!service || (ts_service && ts_service != service))
 		return;
 
 	ts_reset(service);
